#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/time.h>

#include <openssl/err.h>
#include <openssl/rsa.h>
#include <openssl/sha.h>
#include <openssl/evp.h>
#include <openssl/bio.h>
#include <openssl/pem.h>

static EVP_PKEY*
LoadKey(const char* filename) {
  BIO* infile = BIO_new(BIO_s_file());
  if (!infile) {
    ERR_print_errors_fp(stderr);
    return NULL;
  }

  if (BIO_read_filename(infile, filename) <= 0) {
    ERR_print_errors_fp(stderr);
    BIO_free(infile);
    return NULL;
  }

  EVP_PKEY* key = PEM_read_bio_PrivateKey(infile, NULL, NULL, NULL);
  BIO_free(infile);
  if (!key) {
    ERR_print_errors_fp(stderr);
    return NULL;
  }

  return key;
}

static uint64_t
Now() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  uint64_t ret = tv.tv_sec;
  ret *= 1000000;
  ret += tv.tv_usec;
  return ret;
}

int
main() {
  ERR_load_crypto_strings();

  uint8_t digest[20];
  const char message[] = "signed message";

  SHA_CTX sha1;
  SHA1_Init(&sha1);
  SHA1_Update(&sha1, message, sizeof(message) - 1);
  SHA1_Final(digest, &sha1);

  uint8_t sig[512];
  const int fd = open("sig", O_RDONLY);
  ssize_t siglen = read(fd, sig, sizeof(sig));
  close(fd);

  EVP_PKEY* key = LoadKey("key.pem");

  uint64_t start_time = Now();
  const int iterations = 100000;

  for (int i = 0; i < iterations; i++) {
    EVP_PKEY_CTX *ctx = EVP_PKEY_CTX_new(key, NULL);
    EVP_PKEY_verify_init(ctx);
    EVP_PKEY_CTX_set_rsa_padding(ctx, RSA_PKCS1_PADDING);
    EVP_PKEY_CTX_set_signature_md(ctx, EVP_sha1());
    if (1 != EVP_PKEY_verify(ctx, sig, siglen, digest, sizeof(digest))) {
      abort();
    }
  }

  uint64_t end_time = Now();

  printf("%f usec per iteration\n", ((float)(end_time - start_time)) / iterations);

  return 0;
}
